#include <cstdlib>

// ⏕Cu
#include "Gg.h"

// _̐
#define POINTS 100000

// d́i̐j
#define GRAVITY -0.0001f

// ⏕Cu̖O
using namespace gg;

// ϊs
static GgMatrix projectionMatrix;
static GgMatrix viewMatrix;

// gbN{[
static GgTrackball *trackball = 0;

/*
** _Q
**
**    _obt@IuWFNggp`f[^̂߂̊NX
*/
class Drops
{
  // _obt@IuWFNg
  GgBuffer<GLfloat[3]> vertex[4];

public:

  // fXgN^
  virtual ~Drops(void) {}

  // RXgN^
  //    points: _, gravity: d
  Drops(int points = 10000, float gravity = -0.0001f)
  {
    // 
    static const float tmax = sqrt(-4.0f / gravity);
    
    // 
    float *btime = new float[points];
    for (int i = 0; i < points; ++i) btime[i] = tmax * (float)rand() / (float)RAND_MAX;

    // ڂ̒_obt@IuWFNg
    vertex[0].load(GL_ARRAY_BUFFER, points, 0, GL_DYNAMIC_COPY);
    GLfloat (*pos)[3] = (GLfloat (*)[3])glMapBuffer(GL_ARRAY_BUFFER, GL_WRITE_ONLY);

    // _̏ʒu̐ݒ
    for (int i = 0; i < points; ++i)
    {
      float t = btime[i] * btime[i];

      pos[i][0] = 2.0f * (float)rand() / (float)RAND_MAX - 1.0f;
      pos[i][1] = 1.0f + 0.5f * gravity * t;
      pos[i][2] = 2.0f * (float)rand() / (float)RAND_MAX - 1.0f;
    }

    // ڂ̒_obt@IuWFNg̃vÕԂA}bv
    glUnmapBuffer(GL_ARRAY_BUFFER);

    // ڂ̒_obt@IuWFNg
    vertex[1].load(GL_ARRAY_BUFFER, points, 0, GL_DYNAMIC_COPY);
    GLfloat (*vel)[3] = (GLfloat (*)[3])glMapBuffer(GL_ARRAY_BUFFER, GL_WRITE_ONLY);

    // _̏x̐ݒ
    for (int i = 0; i < points; ++i)
    {
      vel[i][0] = 0.0f;
      vel[i][1] = gravity * btime[i];
      vel[i][2] = 0.0f;
    }

    // ڂ̒_obt@IuWFNg̃vÕԂA}bv
    glUnmapBuffer(GL_ARRAY_BUFFER);

    // Oڂ̒_obt@IuWFNg
    vertex[2].load(GL_ARRAY_BUFFER, points, 0, GL_DYNAMIC_COPY);

    // lڂ̒_obt@IuWFNg
    vertex[3].load(GL_ARRAY_BUFFER, points, 0, GL_DYNAMIC_COPY);

    // pɊmۂ̊J
    delete[] btime;
  }

  // Rs[RXgN^
  Drops(const Drops &o)
  {
    for (int i = 0; i < 4; ++i)
    {
      vertex[i] = o.vertex[i];
    }
  }

  // Zq
  Drops &operator=(const Drops &o)
  {
    if (this != &o)
    {
      for (int i = 0; i < 4; ++i)
      {
        vertex[i] = o.vertex[i];
      }
    }

    return *this;
  }

  // _Q`悷葱
  virtual void draw(GLint positionLocation, GLint velocityLocation, GLenum mode = GL_POINTS)
  {
    // Et[
    static int frame = 0;
    
    // ڂ̒_obt@IuWFNg attribute ϐ pos ̒_ɐݒ肷
    glBindBuffer(GL_ARRAY_BUFFER, vertex[0 + frame].buf());
    glEnableVertexAttribArray(positionLocation);
    glVertexAttribPointer(positionLocation, 3, GL_FLOAT, GL_FALSE, 0, 0);

    // ڂ̒_obt@IuWFNg attribute ϐ vel ̒_ɐݒ肷
    glBindBuffer(GL_ARRAY_BUFFER, vertex[1 + frame].buf());
    glEnableVertexAttribArray(velocityLocation);
    glVertexAttribPointer(velocityLocation, 3, GL_FLOAT, GL_FALSE, 0, 0);

    // Oڂ̒_obt@IuWFNg pos ̃^[QbgƂĎw
    glBindBufferBase(GL_TRANSFORM_FEEDBACK_BUFFER, 0, vertex[2 - frame].buf());

    // lڂ̒_obt@IuWFNg^[QbgƂĎw
    glBindBufferBase(GL_TRANSFORM_FEEDBACK_BUFFER, 1, vertex[3 - frame].buf());

    // Transform Feedback Jn
    //GLuint query, primitives_written = 0;
    //glBeginQuery(GL_TRANSFORM_FEEDBACK_PRIMITIVES_WRITTEN, query);
    glBeginTransformFeedback(GL_POINTS);

    // }`̕`
    glDrawArrays(mode, 0, vertex[0].num());

    // Transform Feedback I
    glEndTransformFeedback();
    //glEndQuery(GL_TRANSFORM_FEEDBACK_PRIMITIVES_WRITTEN);
    //glGetQueryObjectuiv(query, GL_QUERY_RESULT, &primitives_written);
    //printf("%d, %d\n", frame, primitives_written);

    // attribute ϐ̒_̐ݒ
    glDisableVertexAttribArray(positionLocation);
    glDisableVertexAttribArray(velocityLocation);

    // Et[̓ւ
    frame = 2 - frame;
  }
};

static Drops *drops = 0;

/*
** _Q̃VF[_
*/
class DropsShader
  : public GgShader
{
  // attribute ϐ̏ꏊ
  GLint positionLocation;
  GLint velocityLocation;
  
  // uniform ϐ̏ꏊ
  GLint transformMatrixLocation;
  GLint gravityLocation;
  
public:
  
  // fXgN^
  ~DropsShader(void) {}
  
  // RXgN^
  DropsShader(void)
  {
    // Transform Feedback  varying ϐ
    const static char *varyings[] = { "pos", "vel" };
    
    // VF[_[̃\[Xt@C̓ǂݍ
    load("DropsShader.vert", "DropsShader.frag", 0, 0, 0, sizeof varyings / sizeof varyings[0], varyings);
    positionLocation = glGetAttribLocation(getProgram(), "position");
    velocityLocation = glGetAttribLocation(getProgram(), "velocity");
    transformMatrixLocation = glGetUniformLocation(getProgram(), "transformMatrix");
    gravityLocation = glGetUniformLocation(getProgram(), "gravity");
  }
  
  // VF[_vOgp
  void use(GgMatrix &projection, GgMatrix &modelview, GLfloat gravity) const
  {
    glUseProgram(getProgram());
    glUniformMatrix4fv(transformMatrixLocation, 1, GL_FALSE, (projection * modelview).get());
    glUniform1f(gravityLocation, gravity);
  }
  
  // attribute ϐ pos ̏ꏊ𓾂
  GLint getPositionLocation(void) const
  {
    return positionLocation;
  }
  
  // attribute ϐ vel ̏ꏊ𓾂
  GLint getVelocityLocation(void) const
  {
    return velocityLocation;
  }
};

static DropsShader *dropsShader = 0;

/*
** n
*/
static void leave(void)
{
  // _f[^̍폜
  delete drops;
  
  // _f[^p̃VF[_̍폜
  delete dropsShader;
  
  // gbN{[̍폜
  delete trackball;
}

/*
** 
*/
static void init(void)
{
  // Q[OtBbNX_̓sɂƂÂ
  ggInit();
  
  // ϊs̐ݒ
  viewMatrix.loadLookat(0.0f, 0.0f, 5.0f, 0.0f, 0.0f, 0.0f, 0.0f, 1.0f, 0.0f);
  
  // gbN{[
  trackball = new GgTrackball;
  
  // _f[^
  drops = new Drops(POINTS, GRAVITY);
  
  // _f[^p̃VF[_
  dropsShader = new DropsShader();
  
  // wiF̐ݒ
  glClearColor(0.0f, 0.1f, 0.3f, 1.0f);
  
  // Bʏ̗L
  glEnable(GL_DEPTH_TEST);
  
  // n
  atexit(leave);
}

/*
** ʕ\
*/
static void display(void)
{
  // fr[ϊs
  GgMatrix modelviewMatrix = viewMatrix * trackball->get();
  
  // ʃNA
  glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
  
  // _Q̕`
  dropsShader->use(projectionMatrix, modelviewMatrix, GRAVITY);
  drops->draw(dropsShader->getPositionLocation(), dropsShader->getVelocityLocation());
  
  // VF[_vO̎w̉
  glUseProgram(0);
  
  // _uobt@O
  glutSwapBuffers();
}

/*
** EBhETCY
*/
static void resize(int w, int h)
{
  // r[|[g̓EBhES
  glViewport(0, 0, w, h);
  
  // eϊs
  projectionMatrix.loadPerspective(GG_DEG_TO_RAD(30.0f), (GLfloat)w / (GLfloat)h, 3.0f, 7.0f);
  
  // gbN{[͈
  trackball->region(w, h);
}

/*
** L[{[h
*/
static void keyboard(unsigned char key, int x, int y)
{
  switch (key)
  {
    // ESC, q, Q ̃^CvŏI
    case '\033':
    case 'q':
    case 'Q':
      exit(0);
    default:
      break;
  }
}

/*
** Aj[V
*/
static void idle(void)
{
  // ʂ̕`ւ
  glutPostRedisplay();
}

/*
** Ă}EX̃{^
*/
static int pressed;

/*
** }EX̃NbN
*/
static void mouse(int button, int state, int x, int y)
{
  switch (pressed = button)
  {
    case GLUT_LEFT_BUTTON:
      if (state == GLUT_DOWN)
      {
        // gbN{[Jn
        trackball->start(x, y);
      }
      else
      {
        // gbN{[~
        trackball->stop(x, y);
      }
      break;
    default:
      break;
  }
}

/*
** }EX̃hbO
*/
static void motion(int x, int y)
{
  switch (pressed)
  {
    case GLUT_LEFT_BUTTON:
      // gbN{[ړ
      trackball->motion(x, y);
      break;
    default:
      break;
  }
}

/*
** CvO
*/
int main(int argc, char *argv[])
{
  glutInit(&argc, argv);
  glutInitDisplayMode(GLUT_RGB | GLUT_DEPTH | GLUT_DOUBLE);
  glutCreateWindow(argv[0]);
  glutDisplayFunc(display);
  glutReshapeFunc(resize);
  glutIdleFunc(idle);
  glutMouseFunc(mouse);
  glutMotionFunc(motion);
  glutKeyboardFunc(keyboard);
  init();
  glutMainLoop();

  return 0;
}
